import { ComponentConstructor } from './components/component';
import { Slide } from './slide';
import { identity } from './types';
export declare type VirchualSettings = {
    slides?: (() => string[]) | null;
    speed?: number;
    easing?: string;
    pagination?: boolean;
    window?: number;
};
export declare class Virchual {
    container: HTMLElement;
    settings: VirchualSettings;
    frame: HTMLElement | null;
    index: number;
    private _slides;
    private _virtualSlidesWindowPrevious;
    private _eventBus;
    private _isMounted;
    private _isBusy;
    private _isDragging;
    private _isEnabled;
    private _pagination;
    constructor(container: HTMLElement, settings?: VirchualSettings);
    /**
     * Return all slides.
     *
     * @param includeClones Whether to include cloned slides or not.
     * @return Slide objects.
     */
    getSlides(includeClones?: boolean): Slide[];
    /**
     * Register component class.
     *
     * @param cls Component class.
     * @param settings Optional settings.
     */
    register<T, U>(cls: ComponentConstructor<T, U>, settings?: U): void;
    /**
     * Mount components.
     */
    mount(): void;
    /**
     * Disable Virchual instance and all user interactions (changing slides)
     */
    disable(): void;
    /**
     * Enable previously disabled Virchual instance. Enable user interaction.
     */
    enable(): void;
    /**
     * Register callback fired on the given event(s).
     *
     * @param events An event name. Use space to separate multiple events.
     *                             Also, namespace is accepted by dot, such as 'resize.{namespace}'.
     * @param handler A callback function.
     * @param elm  Optional. Native event will be listened to when this arg is provided.
     * @param opts Optional. Options for addEventListener.
     */
    on(events: string, handler: identity): void;
    /**
     * Unsubscribe the given event.
     *
     * @param events An event name.
     * @param handler Event handler callback.
     */
    off(events: string, handler: identity): void;
    /**
     * Go to previous slide.
     */
    prev(): void;
    /**
     * Go to next slide.
     */
    next(): void;
    /**
     * Destroy instance, remove DOM events.
     */
    destroy(): void;
    private _getImports;
    /**
     * Return total number of slides.
     *
     * @param includeClones Whether to include cloned slides or not.
     * @return Total number of slides.
     */
    private _getSlidesLength;
    private _goTo;
    /**
     * Initialize slides.
     */
    private _initSlides;
    /**
     * Hydrate existing slides from DOM.
     */
    private _hydrate;
    /**
     * Create clones of slides in case there are less slides than window size.
     */
    private _createClones;
    /**
     * Mount and unmount slides.
     */
    private _mountAndUnmountSlides;
    private _bindEvents;
    /**
     * Called when frame is clicked.
     *
     * @param event A click event.
     */
    private _onClick;
    /**
     * Handle drag event.
     *
     * @param event
     */
    private _onDrag;
    /**
     * Handle dragged event.
     *
     * @param event
     */
    private _onDragEnd;
    /**
     * Move slides to their new position.
     *
     * @param positionX New slide position. Unit (px, %,..) has to be given.
     * @param sign Direction. Either -1 or +1.
     * @param callback Callback function which is called after moving has finished.
     */
    private _move;
}
